/* dirmngr.c - LDAP access
 *	Copyright (C) 2002 Klarlvdalens Datakonsult AB
 *
 * This file is part of DirMngr.
 *
 * DirMngr is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * DirMngr is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include <config.h>
#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <string.h>
#include <assert.h>

#define JNLIB_NEED_LOG_LOGV
#include "crlcache.h"
#include "cert.h"
#include "crlfetch.h"
#include "dirmngr.h"

#include "../assuan/assuan.h" /* needed for the malloc hooks */

/* no i18n needed */
#define N_(a) a
#define _(a) a

#define MAXLEN_CERT 4096
#define GET_PTR( ctx ) ((struct server_control_s*)assuan_get_pointer(ctx))->server_local

/* Data used to associate an Assuan context with local server data */
struct server_local_s {
  ASSUAN_CONTEXT assuan_ctx;
  CrlCache*      crl_cache;
  int message_fd;
  int recursion;
};

static int inquire_cert( ASSUAN_CONTEXT ctx, char* line )
{
  char *value = NULL;
  size_t valuelen; 
  int rc;
  char* buf;
  char* issuer;
  struct server_local_s* ptr = GET_PTR( ctx );
  KsbaCert issuer_cert;

  buf = xmalloc( (strlen(line))+10 * sizeof( char ) );
  strcpy( buf, "SENDCERT");
  /*strcat( buf, line );*/
  rc = assuan_inquire( ctx, buf,
		       &value, &valuelen, MAXLEN_CERT);
  xfree( buf );
  if( rc ) return rc;
  
  issuer_cert = make_cert( value, valuelen );
  if( issuer_cert == NULL ) {
    log_debug( "Error in issuer cert\n");
    return ASSUAN_General_Error;
  }
  issuer = ksba_cert_get_issuer( issuer_cert, 0 );
  xfree( value ); value = NULL;
  if( !issuer ) {
    return ASSUAN_No_CRL_For_Cert; /* FIXME: return better error */    
  }

  /*
  KsbaError ksba_cert_get_crl_dist_point (KsbaCert CERT, int IDX,
					  KsbaName *R_DISTPOINT, KsbaName *R_ISSUER,
					  unsigned int  *R_REASON)
  */
  {    
    int i = 0;
    int hasDistPoint = 0;
    KsbaName dp;
    KsbaName issuername;    
    log_debug( "checking distribution points in %s\n", issuer );
    while( ksba_cert_get_crl_dist_point( issuer_cert, i, &dp, &issuername, NULL ) != -1 ) {
      char* dps;
      char* issuernames;
      
      dps = ksba_name_get_uri( dp, 0 );
      issuernames = ksba_name_get_uri( issuername, 0 );
      
      if( !dps && !issuername ) break;

      fprintf(stderr, "dp=%s, issuername=%s\n", dps, issuernames );
      hasDistPoint = 1;
      ksba_name_release( dp );
      ksba_name_release( issuername );
      ++i;
      //rc = crl_fetch( issuer, &value, &valuelen );
      //ksba_free( issuer );
      rc = crl_fetch( dps, &value, &valuelen );
      if( rc ) {
	xfree( value );
	log_debug( "Error in crl_fetch\n");
	return ASSUAN_No_CRL_For_Cert;
      }
      rc = crl_cache_insert( ptr->crl_cache, value, valuelen );
      xfree( value );
      if( rc ) {
	/* ??? */
	if( rc == CRL_TOO_OLD ) return ASSUAN_CRL_Too_Old;
	else return ASSUAN_No_CRL_For_Cert;    
      }
    }
    if( !hasDistPoint ) {
      rc = crl_fetch_default( issuer, &value, &valuelen );
      if( rc ) {
	xfree( value );
	log_debug( "Error in crl_fetch\n");
	return ASSUAN_No_CRL_For_Cert;
      }
      rc = crl_cache_insert( ptr->crl_cache, value, valuelen );
      xfree( value );	
      if( rc ) {
	/* ??? */
	if( rc == CRL_TOO_OLD ) return ASSUAN_CRL_Too_Old;
	else return ASSUAN_No_CRL_For_Cert;    
      }
    }
  }
  ksba_free( issuer );
  ksba_cert_release( issuer_cert );
  return rc;
}

static int
cmd_isvalid( ASSUAN_CONTEXT ctx, char* line )
{
  struct server_local_s* ptr = GET_PTR( ctx );
  char* ser = line;
  char* saveline = xstrdup( line );
  int rc = 0;

  while( *ser != '\0' && *ser != '.' ) ++ser;
  if( *ser != '.' ) {
    log_debug( "wrong format \"%s\"\n", line);  
    return ASSUAN_General_Error;
  }
    
  /* chop it up */
  *ser = '\0'; ++ser;
  log_debug( "line is  \"%s\"\n", line );

  switch( crl_cache_isvalid( ptr->crl_cache, line, ser ) ) {
  case CRL_CACHE_VALID:
    rc = ASSUAN_No_Error;
    //fprintf(stderr, "VALID\n");
    break;
  case CRL_CACHE_INVALID:
    rc = ASSUAN_Cert_Revoked;
    //fprintf(stderr, "INVALID\n");
    break;
  case CRL_CACHE_DONTKNOW:
    //fprintf(stderr, "DONTKNOW\n");
    if( (rc = inquire_cert( ctx, line )) != ASSUAN_No_Error ) {
      return rc;
    }
    /* do recursive call to get result */
    if( ptr->recursion == 0 ) {
      log_debug( "Calling recursively with \"%s\"\n", saveline );
      ptr->recursion++;
      rc = cmd_isvalid( ctx, saveline );
      ptr->recursion--;
    } else {
      /* we already tried recursively, report an error */
      rc = ASSUAN_No_CRL_For_Cert;
    }
    break;
  }
  /* restore line */
  --ser; *ser = '.';  
  xfree( saveline );
  return rc;
}

/* Copied from gpgsm server.c */
/* note, that it is sufficient to allocate the target string D as
   long as the source string S, i.e.: strlen(s)+1; */
static void
strcpy_escaped_plus (char *d, const unsigned char *s)
{
  while (*s)
    {
      if (*s == '%' && s[1] && s[2])
        {
          s++;
          *d++ = xtoi_2 ( s);
          s += 2;
        }
      else if (*s == '+')
        *d++ = ' ', s++;
      else
        *d++ = *s++;
    }
  *d = 0;
}

static int
cmd_lookup( ASSUAN_CONTEXT ctx, char* line )
{
  char* p;
  STRLIST list, sl;
  int rc;
  int truncated = 0;
  int count = 0;
  char *value = NULL;
  size_t valuelen; 
  int server_count;

  /* break the line down into an STRLIST */
  list = NULL;
  for (p=line; *p; line = p) {
    while (*p && *p != ' ') {
      p++;
    }
    if (*p) {
        *p++ = 0;
    }
    if (*line) {
      sl = xtrymalloc (sizeof *sl + strlen (line));
      if (!sl) {
	free_strlist ( list );
	return ASSUAN_Out_Of_Core;
      }
      sl->flags = 0;
      strcpy_escaped_plus( sl->d, line );
      sl->next = list;
      list = sl;
    }
  }

  for( server_count = 0; dirmngr_opt.ldapservers && dirmngr_opt.ldapservers[server_count].host;
       ++server_count ) {
    /* fetch certificates matching pattern */
    rc = start_cert_fetch( list, &(dirmngr_opt.ldapservers[server_count]) );
    if( rc != ASSUAN_No_Error ) {
      if( rc == ASSUAN_DIRMNGR_Result_Truncated ) {
	truncated = 1;
      } else {
	free_strlist( list );
	return rc;
      }
    }
    do {
      if( value ) {
	xfree( value );
	value = 0;
      }
      rc = fetch_next_cert( &value, &valuelen );
      if( rc != ASSUAN_No_Error ) {
	if( rc == ASSUAN_DIRMNGR_Result_Truncated ) {
	  truncated = 1;
	} else {
	  log_debug("Got error %d\n", rc);
	  return rc;
	}
      }
      if( value ) {
	log_debug("Sending cert\n");
	rc = assuan_send_data( ctx, value, valuelen );      
	if( rc != ASSUAN_No_Error ) {
	  log_debug("Error during assuan_send_data() %d\n", rc);
	  xfree( value );
	  return rc;
	}
	rc = assuan_send_data( ctx, NULL, 0);
	if( rc != ASSUAN_No_Error ) {
	  log_debug("Error during assuan_send_data() %d\n", rc);
	  xfree( value );
	  return rc;
	}
	assuan_write_line (ctx, "END");
	++count;
	if( count >= dirmngr_opt.max_replies ) {
	  break;
	}
      }
    } while( value != NULL );
    end_cert_fetch();
    if( count >= dirmngr_opt.max_replies ) {
      break;
    }
  }
  if( truncated || count >= dirmngr_opt.max_replies ) {
    char str[20]; str[0] = 0;
    snprintf (str, sizeof(str), "%d", count);
    assuan_write_status (ctx, "TRUNCATED", str );    
  }

  free_strlist( list );
  return ASSUAN_No_Error;
}

/* Tell the assuan library about our commands */
static int
register_commands (ASSUAN_CONTEXT ctx)
{
  static struct {
    const char *name;
    int cmd_id;
    int (*handler)(ASSUAN_CONTEXT, char *line);
  } table[] = {
    { "ISVALID",    0,  cmd_isvalid },
    { "LOOKUP",     0,  cmd_lookup },
    { "",     ASSUAN_CMD_INPUT, NULL },
    { "",     ASSUAN_CMD_OUTPUT, NULL },
    { NULL }
  };
  int i, j, rc;

  for (i=j=0; table[i].name; i++)
    {
      rc = assuan_register_command (ctx,
                                    table[i].cmd_id? table[i].cmd_id
                                                   : (ASSUAN_CMD_USER + j++),
                                    table[i].name, table[i].handler);
      if (rc)
        return rc;
    }
  return 0;
}


/* Startup the server */
void
start_command_handler (void)
{
  int rc;
  int filedes[2];
  ASSUAN_CONTEXT ctx;
  struct server_control_s ctrl;

  memset (&ctrl, 0, sizeof ctrl);


  /* For now we use a simple pipe based server so that we can work
     from scripts.  We will later add options to run as a daemon and
     wait for requests on a Unix domain socket */
  filedes[0] = 0;
  filedes[1] = 1;
  rc = assuan_init_pipe_server (&ctx, filedes);
  if (rc)
    {
      log_error ("failed to initialize the server: %s\n",
                 assuan_strerror(rc));
      dirmgr_exit (2);
    }

  rc = register_commands (ctx);
  if (rc)
    {
      log_error ("failed to the register commands with Assuan: %s\n",
                 assuan_strerror(rc));
      dirmgr_exit (2);
    }

  ctrl.server_local = xcalloc (1, sizeof *ctrl.server_local);
  ctrl.server_local->assuan_ctx = ctx;
  ctrl.server_local->crl_cache = crl_cache_init();
  ctrl.server_local->message_fd = -1;
  ctrl.server_local->recursion = 0;

  assuan_set_pointer (ctx, &ctrl);

  if ((dirmngr_opt.debug & 1024))
    assuan_set_log_stream (ctx, log_get_stream ());

  for (;;) {
    rc = assuan_accept (ctx);
    if (rc == -1) {
      break;
    }
    else if (rc) {
      log_info ("Assuan accept problem: %s\n", assuan_strerror (rc));
      break;
    }

    rc = assuan_process (ctx);
    if (rc) {
      log_info ("Assuan processing failed: %s\n", assuan_strerror (rc));
      continue;
    }
  }

  crl_cache_deinit(  ctrl.server_local->crl_cache );
  assuan_deinit_server (ctx);
}
